//	キャラクタLCD I2C化アダプタ：マスター側ライブラリ
//	『昼夜逆転』工作室	http://jsdiy.webcrow.jp
//	2017/08	初版
//	2017/09	・SendByteXXX()の戻り値（正否結果）を廃止した。
//			・I2CLCD_PutWord()をSendByteBlock()呼出からSendByteP2()に変更。

#include <avr/io.h>
#include <avr/pgmspace.h>
#include "I2cCharLcdLib.h"
#include "../twilib.h"

#define HiByte(w)	(((w) >> 8) & 0xFF)
#define LoByte(w)	((w) & 0xFF)

#define SLA_W_BIT	0
#define SLA_R_BIT	1

static uint8_t	SlaveAddress = I2CLCD_DEFAULT_SLA;

//コントロールバイトと1個のデータバイトを送信する
static	void	SendByteP1(uint8_t ctrlByte, uint8_t dataByte)
{
	TWI_Start();
	TWI_SlaveAddrW(SlaveAddress);
	TWI_WriteByte(ctrlByte);
	TWI_WriteByte(dataByte);
	TWI_Stop();
}

//コントロールバイトと2個のデータバイトを送信する
static	void	SendByteP2(uint8_t ctrlByte, uint8_t dataByte1, uint8_t dataByte2)
{
	TWI_Start();
	TWI_SlaveAddrW(SlaveAddress);
	TWI_WriteByte(ctrlByte);
	TWI_WriteByte(dataByte1);
	TWI_WriteByte(dataByte2);
	TWI_Stop();
}

//コントロールバイトとデータバイトの配列を送信する
static	void	SendByteBlock(uint8_t ctrlByte, const uint8_t* dataBytes, uint8_t length)
{
	uint8_t	i;
	
	TWI_Start();
	TWI_SlaveAddrW(SlaveAddress);
	for (i = 0; i < length; i++)
	{
		TWI_WriteByte(ctrlByte);
		TWI_WriteByte(dataBytes[i]);
	}
	TWI_Stop();
}

//操作対象のLCDデバイスを初期化する
//・本デバイスはスレーブアドレスを変えてI2Cバスに複数個接続することができる。
//	それぞれのデバイスに対してこの関数を呼び出す必要がある。
//・I2Cバスに本デバイスを1個しか接続していない場合、この関数は1回呼び出せば十分。
void	I2CLCD_Init(uint8_t slaveAddress, uint8_t columnSize, uint8_t rowSize)
{
	I2CLCD_SlaveSelect(slaveAddress);
	SendByteP2(LCD_FUNC_INIT, columnSize, rowSize);
}

//操作対象のLCDデバイスを切り替える
//・本デバイスはスレーブアドレスを変えてI2Cバスに複数個接続することができる。
//	その場合に、どのLCDデバイスを操作するのか、その都度この関数で切り替える。
//・I2Cバスに本デバイスを1個しか接続していない場合、この関数を呼び出す場面はない。
void	I2CLCD_SlaveSelect(uint8_t slaveAddress)
{
	SlaveAddress = slaveAddress;
}

//HD44780レベルのコマンドを送信する（詳細はHD44780のデータシートのコマンド一覧を参照）
void	I2CLCD_SendCmd(uint8_t cmd)
{
	SendByteP1(RS_CMD, cmd);
}

//HD44780レベルのデータを送信する（詳細はHD44780のデータシートのコマンド一覧を参照）
void	I2CLCD_SendData(uint8_t data)
{
	SendByteP1(RS_DATA, data);
}

//行表示の動作をセットする
void	I2CLCD_SetLineMode(uint8_t lineMode, uint8_t cursorFlag)
{
	SendByteP2(LCD_FUNC_SETLINEMODE, lineMode, cursorFlag);
}

//文字の表示位置をセットする
void	I2CLCD_SetPos(uint8_t row_y, uint8_t column_x)
{
	SendByteP2(LCD_FUNC_SETPOS_XY, column_x, row_y);
}

//LCD画面を消す
void	I2CLCD_Clear()
{
	I2CLCD_SendCmd(LCD_CMD_CLEAR_DISPLAY);
}

//1文字表示する
void	I2CLCD_PutChar(char cc)
{
	SendByteP1(LCD_FUNC_PUTCHAR, cc);
}

//文字列を表示する
//・一度に送信できる文字数は (UINT8_MAX - 1) 文字まで。
void	I2CLCD_PutString(const char* str)
{
	uint8_t length;
	for (length = 0; str[length] != '\0'; length++);
	SendByteBlock(LCD_FUNC_PUTCHAR, (uint8_t*)str, length);
}

//プログラム領域に格納された文字列を表示する
//・一度に送信できる文字数は (UINT8_MAX - 1) 文字まで。
void	I2CLCD_PutStringP(const char* strpgm)
{
	uint8_t cc;
	
	TWI_Start();
	TWI_SlaveAddrW(SlaveAddress);
	while ((cc = pgm_read_byte(strpgm++)) != '\0')
	{
		TWI_WriteByte(LCD_FUNC_PUTCHAR);
		TWI_WriteByte(cc);
	}
	TWI_Stop();
}

//符号あり整数を表示する（10進数）	※8bit整数/16bit整数兼用
void	I2CLCD_PutInt(int16_t n)
{
	SendByteP2(LCD_FUNC_PUTINT, HiByte(n), LoByte(n));
}

//符号なし整数を表示する（10進数）	※8bit整数/16bit整数兼用
void	I2CLCD_PutUInt(uint16_t n)
{
	SendByteP2(LCD_FUNC_PUTUINT, HiByte(n), LoByte(n));
}

//0詰め2桁で16進数を表示する	※接頭辞'0x'や接尾辞'h'は付かない
void	I2CLCD_PutByte(uint8_t n)
{
	SendByteP1(LCD_FUNC_PUTBYTE, n);
}

//0詰め4桁で16進数を表示する	※接頭辞'0x'や接尾辞'h'は付かない
void	I2CLCD_PutWord(uint16_t n)
{
	SendByteP2(LCD_FUNC_PUTWORD, HiByte(n), LoByte(n));
}

//ユーザー文字を定義する（詳細はHD44780のデータシートのコマンド一覧を参照）
//id:	0-7（キャラクタコード）
//dots:	uint8_t[8]（5x8のドットパターン）
void	I2CLCD_DefChar(uint8_t id, const uint8_t* dots)
{
	uint8_t	i;
	for (i = 0; i < 8; i++)
	{
		I2CLCD_SendCmd(LCD_CMD_SET_CGRAM_ADRS | (id << 3) | i);
		I2CLCD_SendData(dots[i]);
	}
}
